//-------------------------------------------------------------------------------------------------
// 0.0 -  What these 2 guys built that I worked from
//        Graphic LCD with Toshiba T6963 controller
//        Copyright (c) Radosaw Kwiecie, 2007r
//        http://en.radzio.dxp.pl/t6963/
//        Compiler : avr-gcc
//        Modified By -Gil- to work on Arduino easily : http://domoduino.tumblr.com/
// 0.1 -  First Hack By rbrsidedn
//        Invocable class T6963
//        Commands moved to T6963_commands.h
// 0.2 -  rewritten for Arduino Due by Marvin G. / apr 3 2013
//        added basic Graphics functions: Rectangle, Line, Circle
//        make sure coordinates are on the graphics area!
//        added variable delay to optimize for individual display driver speed.
//        Try 24 and check if garbage apears or not. You may need to increase the delay.
//-------------------------------------------------------------------------------------------------
#include "Arduino.h"
#include "T6963due.h"
#include "T6963due_Commands.h"
//-------------------------------------------------------------------------------------------------
//
// Delay function
//
//-------------------------------------------------------------------------------------------------
void T6963::n_delay(void)
{
  volatile byte i;
  for(i = 0; i < _DELAY; i++)
  {
    asm("nop");
  }
}
//-------------------------------------------------------------------------------------------------
//
// Constructor
//
//-------------------------------------------------------------------------------------------------

T6963::T6963(int pixHoriz,int pixVert,int fontWidth,int delay){
  _DELAY = delay;
  GLCD_NUMBER_OF_LINES = pixVert;
  GLCD_PIXELS_PER_LINE = pixHoriz;
  _FW = fontWidth;
  _GA = pixHoriz / fontWidth;	//Supercedes GLCD_GRAPHIC_AREA
  _TA = pixHoriz / fontWidth;	//Supercedes GLCD_TEXT_AREA
  sizeGA = _GA*pixVert;		//Supercedes GLCD_GRAPHIC_SIZE
  sizeTA = _TA*pixVert/8;	        //Supercedes GLCD_TEXT_SIZE
  setTH(0);
  setGH(0);
}
void T6963::setTH(unsigned int addr){
  _TH=addr;
} //expose _TH

unsigned int T6963::getTH(){
  return _TH;
}	//return _TH

void T6963::setGH(unsigned int addr){
  if(addr == _TH){
    _GH=_TH+sizeTA;
  }
  else{
    _GH=addr;
  }
}
unsigned int T6963::getGH(){
  return _GH;
}



//-------------------------------------------------------------------------------------------------
//
// Ports intalization
//
//-------------------------------------------------------------------------------------------------
void T6963::InitalizeInterface(void){
  //Activates the PIO clock on portD, detach perepherals
  pinMode(25,INPUT);
  pinMode(14,INPUT);
  pinMode(15,INPUT);
  //out mode
  GLCD_DATA_OUT;  //all outputs
  GLCD_CTRL_OUT;  //also outputs
  GLCD_CTRL_HIGH = GLCD_CTRL_ALL;  //all control pins high
}
//-------------------------------------------------------------------------------------------------
//
// Reads dispay status
//
//-------------------------------------------------------------------------------------------------
byte T6963::checkStatus(void){
  uint8_t tmp;
  //in mode
  GLCD_DATA_IN;  //all inputs
  GLCD_CTRL_LOW = GLCD_CTRL_CHECK;
  n_delay();
  tmp = GLCD_DATA_READ;
  //out mode
  GLCD_DATA_OUT;  //all outputs
  GLCD_CTRL_HIGH = GLCD_CTRL_CHECK;  //reset rd and ce
  return tmp;
}
//-------------------------------------------------------------------------------------------------
//
// Writes instruction
//
//-------------------------------------------------------------------------------------------------
void T6963::writeCommand(byte command){
  while(!(checkStatus()&0x03));
  // GLCD_DATA_PORT = command;
  GLCD_DATA_SEL = GLCD_DATA_MASK; //Select output channels
  GLCD_DATA_UNSEL = ~GLCD_DATA_MASK;
  GLCD_DATA_WRITE = command;

  GLCD_CTRL_LOW = GLCD_CTRL_WRCMD;
  n_delay();
  GLCD_CTRL_HIGH = GLCD_CTRL_WRCMD;
}

//-------------------------------------------------------------------------------------------------
//
// Writes data
//
//-------------------------------------------------------------------------------------------------
void T6963::WriteData(byte data){
  while(!(checkStatus()&0x03));
  // GLCD_DATA_PORT = data;
  GLCD_DATA_SEL = GLCD_DATA_MASK; //Select output channels
  GLCD_DATA_UNSEL = ~GLCD_DATA_MASK;
  GLCD_DATA_WRITE = data;

  GLCD_CTRL_LOW = GLCD_CTRL_WRDATA;
  n_delay();
  GLCD_CTRL_HIGH = GLCD_CTRL_WRDATA;
}
//-------------------------------------------------------------------------------------------------
//
// Reads data
//
//-------------------------------------------------------------------------------------------------
byte T6963::ReadData(void){
  byte tmp;
  while(!(checkStatus()&0x03));
  GLCD_DATA_IN;

  GLCD_CTRL_LOW = GLCD_CTRL_RDDATA;
  n_delay();

  tmp = GLCD_DATA_READ;

  GLCD_CTRL_HIGH = GLCD_CTRL_RDDATA;
  GLCD_DATA_OUT;
  return tmp;
}
//-------------------------------------------------------------------------------------------------
//
// Sets address pointer for display RAM memory
//
//-------------------------------------------------------------------------------------------------
void T6963::SetAddressPointer(unsigned int address){
  WriteData(address & 0xFF);  //Low byte
  WriteData(address >> 8);    //high byte
  writeCommand(T6963_SET_ADDRESS_POINTER);
}
//-------------------------------------------------------------------------------------------------
//
// Clears text area of display RAM memory
//
//-------------------------------------------------------------------------------------------------
void T6963::clearText(){
  SetAddressPointer(_TH);
  for(int i = 0; i < sizeTA; i++){
    WriteDisplayData(0);
  }
}
//-------------------------------------------------------------------------------------------------
//
// Clears characters generator area of display RAM memory
//
//-------------------------------------------------------------------------------------------------
void T6963::clearCG(){
  unsigned int i;
  SetAddressPointer(GLCD_EXTERNAL_CG_HOME);
  for(i = 0; i < 256 * 8; i++){
    WriteDisplayData(0);
  }
}
//-------------------------------------------------------------------------------------------------
//
// Clears graphics area of display RAM memory
//
//-------------------------------------------------------------------------------------------------
void T6963::clearGraphic(){
  SetAddressPointer(_GH);
  for(unsigned int i = 0; i < sizeGA; i++){
    WriteDisplayData(0x00);
  }
}
//-------------------------------------------------------------------------------------------------
//
// Writes a single character (ASCII code) to display RAM memory
//
//-------------------------------------------------------------------------------------------------
void T6963::WriteChar(char charCode)
{
  WriteDisplayData(charCode - 32);
}
//-------------------------------------------------------------------------------------------------
//
// Writes null-terminated string to display RAM memory
//
//-------------------------------------------------------------------------------------------------
void T6963::WriteString(char * string){
  while(*string){
    WriteChar(*string++);
  }
}
//-------------------------------------------------------------------------------------------------
//
// Sets display coordinates
//
//-------------------------------------------------------------------------------------------------
void T6963::TextGoTo(unsigned char x, unsigned char y){
  unsigned int address;
  address = _TH +  x + (_TA * y);
  SetAddressPointer(address);
}
//-------------------------------------------------------------------------------------------------
//
// Writes single char pattern to character generator area of display RAM memory
//
//-------------------------------------------------------------------------------------------------
void T6963::DefineCharacter(unsigned char charCode, unsigned char * defChar){
  unsigned int address;
  unsigned char i;
  address = GLCD_EXTERNAL_CG_HOME + (8 * charCode);
  SetAddressPointer(address);
  for(i = 0; i < 8 ; i++){
    WriteDisplayData(*(defChar + i));
  }
}

//-------------------------------------------------------------------------------------------------
//
// Set (if color==1) or clear (if color==0) pixel on screen
//
//-------------------------------------------------------------------------------------------------
void T6963::setPixel(unsigned char x, unsigned char y, unsigned char color){
  unsigned char tmp;
  unsigned int address;
  address = _GH + x/_FW + _GA*y;
  SetAddressPointer(address);
  writeCommand(T6963_DATA_READ_AND_NONVARIABLE);
  tmp = ReadData();
  if(color){
    tmp |= (1 <<  (_FW - 1 - (x % _FW)));
  }
  else{
    tmp &= ~(1 <<  (_FW - 1 - (x % _FW)));
  }
  WriteDisplayData(tmp);
}
//-------------------------------------------------------------------------------------------------
//
// Writes display data and increment address pointer
//
//-------------------------------------------------------------------------------------------------
void T6963::WriteDisplayData(byte x){
  WriteData(x);
  writeCommand(T6963_DATA_WRITE_AND_INCREMENT);
}
//-------------------------------------------------------------------------------------------------
//
// Sets graphics coordinates
//
//-------------------------------------------------------------------------------------------------
void T6963::GraphicGoTo(byte x, byte y){
  unsigned int address;
  address = _GH + (x / _FW) + (_GA * y);
  SetAddressPointer(address);
}



//-------------------------------------------------------------------------------------------------
//
// Displays bitmap from program memory
//
//-------------------------------------------------------------------------------------------------
/*
void T6963::Bitmap(byte * bitmap, byte x, byte y, byte width, byte height){
 byte i, j;
 for(j = 0; j < height; j++){
 GraphicGoTo(x, y + j);
 for(i = 0; i < width/_FW; i++){
 WriteDisplayData(pgm_read_byte(bitmap + i + (_GA * j)));
 }
 }
 }
 */
byte T6963::setMode(char _mode, char _CG){
  byte tmp = T6963_MODE_SET;
  if(_mode=='X' || _mode =='x' || _mode=='^')  //XOR
    tmp |= 1;
  else if(_mode == 'A' || _mode=='&' ||_mode=='a')   //AND
    tmp |= 3;
  else if(_mode == 'T'||_mode=='t')   //Text attribute
    tmp |= 4;
  else{
    tmp |= 0; //OR mode default
  }
  if(_CG =='E'||_CG=='e'){
    tmp |=8;
  }
  else{
    tmp |=0;
  }
  writeCommand(tmp);
  return tmp;
}

byte T6963::clearDispMode(){
  writeCommand(T6963_DISPLAY_MODE);
}
byte T6963::setDispMode(boolean _text,boolean _graphics, boolean _cursor, boolean _blink){
  byte tmp=T6963_DISPLAY_MODE;
  if(_text){
    tmp |= 0b0100;
  }
  if(_graphics){
    tmp |= 0b1000;
  }
  if(_cursor){
    tmp |= 0b0010;
  }
  if(_blink){
    tmp |= 0b0001;
  }
  writeCommand(tmp);
  return tmp;
}


//-------------------------------------------------------------------------------------------------
//
// Display initalization
//
//-------------------------------------------------------------------------------------------------
void T6963::Initialize(void){
  //Set up data and control ports
  InitalizeInterface();
  //reset LCD, should be able to have hardware solution
  GLCD_CTRL_LOW = GLCD_CTRL_RESET;
  n_delay();
  GLCD_CTRL_HIGH = GLCD_CTRL_RESET;
  //Set Graphics Home Address
  WriteData(_GH & 0xFF);
  WriteData(_GH >> 8);
  writeCommand(T6963_SET_GRAPHIC_HOME_ADDRESS);
  //Set Graphics Area
  WriteData(_GA);
  WriteData(0x00);
  writeCommand(T6963_SET_GRAPHIC_AREA);
  //Set Text home address
  WriteData(_TH & 0xFF);
  WriteData(_TH >> 8);
  writeCommand(T6963_SET_TEXT_HOME_ADDRESS);
  //Set Text Area
  WriteData(_TA);
  WriteData(0x00);
  writeCommand(T6963_SET_TEXT_AREA);
  //Set Internal CGRAM address
  WriteData(15);
  WriteData(0x00);
  writeCommand(T6963_SET_OFFSET_REGISTER);

  setDispMode(true,true,true,true);

  setMode('O','I');

  clearText();
  clearGraphic();
  delay(100);

}
//-------------------------------------------------------------------------------------------------
//
// Displays buffer
//
//-------------------------------------------------------------------------------------------------
void T6963::DispBuff(byte * buff){
   SetAddressPointer(_GH);
   for(unsigned int i = 0; i < sizeGA; i++)
     WriteDisplayData(buff[i]);
}
//-------------------------------------------------------------------------------------------------
//
// Basic Graphics, unoptimized!
//
//-------------------------------------------------------------------------------------------------

void T6963::Circle(byte cx, byte cy, byte radius, byte color){
  int x = radius, y = 0, xchange, ychange, radiusError;
  xchange = 1 - radius*2;
  ychange = 1;
  radiusError = 0;
  while(x >= y)
  {
    setPixel(cx+x, cy+y, color);
    setPixel(cx-x, cy+y, color);
    setPixel(cx-x, cy-y, color);
    setPixel(cx+x, cy-y, color);
    setPixel(cx+y, cy+x, color);
    setPixel(cx-y, cy+x, color);
    setPixel(cx-y, cy-x, color);
    setPixel(cx+y, cy-x, color);
    y++;
    radiusError += ychange;
    ychange += 2;
    if ( 2*radiusError + xchange > 0 )
    {
      x--;
      radiusError += xchange;
      xchange += 2;
    }
  }

}
void T6963::Line(byte X1, byte Y1,byte X2,byte Y2,byte color)
{
int CurrentX, CurrentY, Xinc, Yinc, Dx, Dy, TwoDx, TwoDy, TwoDxAccumulatedError, TwoDyAccumulatedError;

Dx = (X2-X1);
Dy = (Y2-Y1);

TwoDx = Dx + Dx;
TwoDy = Dy + Dy;

CurrentX = X1;
CurrentY = Y1;

Xinc = 1;
Yinc = 1;

if(Dx < 0)
  {
  Xinc = -1;
  Dx = -Dx;
  TwoDx = -TwoDx;
  }

if (Dy < 0)
  {
  Yinc = -1;
  Dy = -Dy;
  TwoDy = -TwoDy;
  }

setPixel(X1,Y1, color);

if ((Dx != 0) || (Dy != 0))
  {
  if (Dy <= Dx)
    {
    TwoDxAccumulatedError = 0;
    do
	  {
      CurrentX += Xinc;
      TwoDxAccumulatedError += TwoDy;
      if(TwoDxAccumulatedError > Dx)
        {
        CurrentY += Yinc;
        TwoDxAccumulatedError -= TwoDx;
        }
       setPixel(CurrentX,CurrentY, color);
       }while (CurrentX != X2);
     }
   else
      {
      TwoDyAccumulatedError = 0;
      do
	    {
        CurrentY += Yinc;
        TwoDyAccumulatedError += TwoDx;
        if(TwoDyAccumulatedError>Dy)
          {
          CurrentX += Xinc;
          TwoDyAccumulatedError -= TwoDy;
          }
         setPixel(CurrentX,CurrentY, color);
         }while (CurrentY != Y2);
    }
  }
}
void T6963::Rectangle(byte x, byte y, byte b, byte a, byte color)
{
  byte j;
  for (j = 0; j < a; j++) {
		setPixel(x, y + j, color);
		setPixel(x + b - 1, y + j, color);
	}
  for (j = 0; j < b; j++)	{
		setPixel(x + j, y, color);
		setPixel(x + j, y + a - 1, color);
	}
}
