/*-------------------------------------------------------------------------------------------------
// 0.0 -  What these 2 guys built that I worked from
//        Graphic LCD with Toshiba T6963 controller
//        Copyright (c) Radosaw Kwiecie, 2007r
//        http://en.radzio.dxp.pl/t6963/
//        Compiler : avr-gcc
//        Modified By -Gil- to work on Arduino easily : http://domoduino.tumblr.com/
// 0.1 -  Invocable class T6963
//        Commands moved to T6963_commands.h
// 0.2 -  rewritten for Arduino Due by Marvin G. / apr 3 2013
//        added basic Graphics functions: Rectangle, Line, Circle
//        make sure coordinates are on the graphics area!
//        added variable delay to optimize for individual display driver speed.
//        Try 24 and check if garbage apears or not. You may need to increase the delay.
//-----------------------------------------------------------------------------------------------*/
#ifndef T6963_h
#define T6963_h

#include "Arduino.h"
#include "inttypes.h"
#include "T6963due_Commands.h"

// data port  (Bus 1-8 goes to pins 25, 26, 27, 28, 14, 15, 29, 11)
#define GLCD_DATA_READ      REG_PIOD_PDSR & 0x000000FF
#define GLCD_DATA_WRITE     REG_PIOD_ODSR
#define GLCD_DATA_OUT       REG_PIOD_OER = 0x000000FF
#define GLCD_DATA_IN        REG_PIOD_ODR = 0x000000FF
#define GLCD_DATA_MASK		0x000000FF
#define GLCD_DATA_SEL   	REG_PIOD_OWER
#define GLCD_DATA_UNSEL   	REG_PIOD_OWDR


// control port (Pins: WR: 34 | RD: 35 | CE: 36 (Should be able to NOR this with WR and RD) | CD: 37 | RESET: 38 | FS: 39, but hardware solution recommended)
#define GLCD_CTRL_OUT		REG_PIOC_OER = 0x000000FC
#define GLCD_CTRL_HIGH      REG_PIOC_SODR
#define GLCD_CTRL_LOW       REG_PIOC_CODR

// control signals
#define GLCD_CTRL_ALL       0b11111100  //ALL
#define GLCD_CTRL_CHECK     0b00011000  //RD, CE
#define GLCD_CTRL_WRCMD     0b00010100  //WR, CE
#define GLCD_CTRL_WRDATA    0b00110100  //WR, CE, CD
#define GLCD_CTRL_RDDATA    0b00111000  //RD, CE, CD
#define GLCD_CTRL_RESET     0b01000000  //RESET


// display properties
#define GLCD_OFFSET_REGISTER				2
#define GLCD_EXTERNAL_CG_HOME				(GLCD_OFFSET_REGISTER << 11)



class T6963{
public:
  T6963(int pixHoriz,int pixVert,int fontWidth, int delay); //Class
  void n_delay(void);
  void InitalizeInterface(void);
  void writeCommand(byte);
  void WriteData(byte);
  byte ReadData(void);
  void clearText(void);
  void clearCG(void);
  void clearGraphic(void);
  void WriteChar(char ch);
  void WriteString(char * str);
  void TextGoTo(byte, byte);
  void GraphicGoTo(byte x, byte y);
  void DefineCharacter(unsigned char, unsigned char *);
  void Initialize(void);
  void setPixel(byte, byte, unsigned char);
  void WriteDisplayData(byte);
  byte checkStatus(void);
  void SetAddressPointer(unsigned int address);
  byte setMode(char, char);
  byte clearDispMode();
  byte setDispMode(boolean _text,boolean _graphics, boolean _cursor, boolean _blink);

  void DispBuff(byte * buff);

  void Circle(byte cx, byte cy, byte radius, byte color);
  void Line(byte X1, byte Y1, byte X2, byte Y2, byte color);
  void Rectangle(byte x, byte y, byte b, byte a, byte color);
  
  void setTH(unsigned int addr);
  unsigned int getTH();
  void setGH(unsigned int addr);
  unsigned int getGH();
  unsigned int sizeGA;	//(GLCD_GRAPHIC_AREA * GLCD_NUMBER_OF_LINES)
  unsigned int sizeTA;	//(GLCD_TEXT_AREA * (GLCD_NUMBER_OF_LINES/8))

private:
  unsigned int GLCD_NUMBER_OF_LINES;	//Veritical Resolution, Lines
  unsigned int GLCD_PIXELS_PER_LINE;	//Horizontal Resolution, pixels per line
  unsigned int _FW;		//Font Width
  unsigned int _GH;		//Graphics home
  unsigned int _TH;		//Text Home
  unsigned int _GA;		//(GLCD_PIXELS_PER_LINE / GLCD_FONT_WIDTH)
  unsigned int _TA;		//(GLCD_PIXELS_PER_LINE / GLCD_FONT_WIDTH)
  unsigned int _DELAY;  //Wait time for display reaction
};


#endif

